//----------------------------------------------------------------------------
//~Author: Fabio Arciniegas A.
//~Name: DOMImageCrawler
/* Description: A DOM Visitor that takes all the images that correspond to a
              particular keyword, and saves them to disk.
              This Visitor only works with XML with the following DTD:

<!ENTITY % metadata "author,date,keywords*">

<!-- An Article is composed by metadata and paragraphs with images -->
<!ELEMENT article  (%metadata;,para+)>
<!ELEMENT para     (#PCDATA|image|a)*>
<!ELEMENT a        (#PCDATA)>
<!ELEMENT author   (#PCDATA)>
<!ELEMENT date     (#PCDATA)>
<!ELEMENT keywords (#PCDATA)>
<!ATTLIST a href      CDATA   #REQUIRED> <!--simple links to ABSOLUTE URLs -->
<!ELEMENT image    (%metadata;)>
<!ATTLIST image    src    CDATA   #REQUIRED>    

*/
//
//Version $Id$
//~Important: See distribution note at the end of this file.

//----------------------------------------------------------------------------
//             This code has been formated to fit in 80 columns
//----------------------------------------------------------------------------
package com.faa.dom.tutorial;

import java.lang.   Thread;
import java.util.   Vector;
import java.util.   Hashtable;
import java.util.   Enumeration;
import java.lang.   Integer;
import org.w3c.dom. Attr;
import org.w3c.dom. Node;
import org.w3c.dom. Element;
import org.w3c.dom. NodeList;
import org.w3c.dom. Document;

import com.faa.patterns.Observer. Observable;
import com.faa.patterns.Observer. Observer;

import java.net. MalformedURLException;
import java.io.  IOException;



/**
    A DOM Visitor that takes all the images that match a particular keyword,
    and saves them to disk.
    This Visitor only works with XML with the following DTD:
<pre>
&lt;!ENTITY % metadata "author,date,keywords*">

&lt;!-- An Article is composed by metadata and paragraphs with images -->
&lt;!ELEMENT article  (%metadata;,para+)>
&lt;!ELEMENT para     (#PCDATA|image|a)*>
&lt;!ELEMENT a        (#PCDATA)>
&lt;!ELEMENT author   (#PCDATA)>
&lt;!ELEMENT date     (#PCDATA)>
&lt;!ELEMENT keywords (#PCDATA)>
&lt;!ATTLIST a href      CDATA   #REQUIRED> &lt;!--simple links to ABSOLUTE URLs -->
&lt;!ELEMENT image    (%metadata;)>
&lt;!ATTLIST image    src    CDATA   #REQUIRED>    
</pre>
*/

public class DOMImageCrawler extends Thread implements Observable
{


// ******************************************         Constructors / Destructors

/**
   Construct a new DOMImageCrawler, starting with the given parameters.
 * @param depth the depth of the search
 * @param initialURL initial place to start the search
 * @param keywords the given keywords
 */
    public DOMImageCrawler(int newDepth,String initialURL, 
                           Vector newKeywords,String newTargetDir)
    {
	setDepth(newDepth);
	setKeywords(newKeywords);
	currentLevel = 0;
	toVisit.put(initialURL,new Integer(0));
	targetDir = newTargetDir;
    }
    
    public  void setDepth(int newDepth)
    {
	depth = newDepth;
    }

    public  void setKeywords(Vector newKeywords)
    {
	keywords = newKeywords;
    }
    
// ******************************************               DOM Specific Methods

  /** 
      Visit the tree, watching for image nodes with the appropriate keywords.
      @param n the current node
   */
    public  void visitNode(Node n)
	throws MalformedURLException, IOException
    {
	if(n.getNodeType() == Node.ELEMENT_NODE)
	{
	    // Check if this is a keywords node with one of our desired strings
	    if(n.getNodeName().equals("keywords"))
	    {
		NodeList children = n.getChildNodes();
		boolean found = false;
		if(children != null)
		{
		    int len = children.getLength();
		    for(int i = 0; i < len ; i++)
			for(int j = 0;j < keywords.size() && !found;j++)
			    found = children.item(i).getNodeValue().indexOf(
                                     (String)keywords.elementAt(j)) != -1; 
		}

		// System.out.println("there is hope here");
		// looking for a substring is not optimal here,but will suffice
		if(found && n.getParentNode().getNodeName().equals("image")) 
		    saveImage((Element)n.getParentNode());
	    }

	    // Look for new links to explore
	    if(n.getNodeName().equals("a") && currentLevel < depth)
	    {
		String href = ((Element)n).getAttribute("href");
		if(!visited.contains(href))
		    toVisit.put(href,new Integer(currentLevel+1));
		for(int i = 0;i < observers.size(); i++)
		    ((Observer)observers.elementAt(i)).notify(this);

	    }
	    NodeList children = n.getChildNodes();
	    if(children != null)
	    {
		int len = children.getLength();
		for(int i = 0; i < len ; i++)
		    visitNode(children.item(i));
	    }
	}
    }

    /** 
       Extract the name of the image from the DOM Node and save it to disk
      @param n the paragraph object containing the image
   */
    public  void saveImage(Element n) 
    {
	    String imageName = n.getAttribute("src");
	    System.out.println(imageName);
	    
	    try 
	    {
		URLSaver saver = new URLSaver(currentDocument,targetDir);
		saver.saveURL(imageName);
		imagesOnThisDocument++; //ToFix: this assummes it all went well. 
		loadedImagesCount++;    
	    }
	    catch(MalformedURLException e)
	    {
		System.out.println("Unable to save image(bad URL) " + imageName);
	    }
	    catch(IOException e)
	    {
		System.out.println("Unable to save image(IO) " + imageName);
	    }

	    
    }
    
// ******************************************                             Thread
  /** 
      Iterate through the nodes to visit, passing each new document to the
      visitor method, so it can be scanned
   */
 
    public  void run()
    {
	while(toVisit.size() > 0)
	{
	    Enumeration keys = toVisit.keys();
	    String key = (String) keys.nextElement();
	    currentLevel = ((Integer)toVisit.get(key)).intValue();
            toVisit.remove(key);


            currentDocument = key; //FIFO
	    imagesOnThisDocument = 0;

	    try {
		dom.DOMParserWrapper parser = 
		    (dom.DOMParserWrapper)
                    Class.forName(DEFAULT_PARSER_NAME).newInstance();
		Document document = parser.parse(currentDocument);
		visitNode(document.getDocumentElement());
		visited.put(currentDocument,new Integer(imagesOnThisDocument));
		//Now, notify the observers that this model changed
		for(int i = 0;i < observers.size(); i++)
		    ((Observer)observers.elementAt(i)).notify(this);
	    }
	    catch (Exception e) {
		e.printStackTrace(System.err);
	    }
	}
	for(int i = 0;i < observers.size(); i++)
	((Observer)observers.elementAt(i)).notify(this);
    }

// ******************************************               Observable Behavior
/**
   Add an observer
   @param observer the observer object to be added.
 */
    public  void addObserver(Observer observer)
    {
	observers.add(observer);
	observer.notify(this);
    }
    
  /** 
      Remove an object from the list of of observers (if it is not there,
      do nothing)
      @param observer
   */
    public  void removeObserver(Observer observer)
    {
	observers.remove(observer);
    }
    
// ******************************************                     Simple getters    

    public  int getCurrentLevel()
    {
	return currentLevel;
    }

    public  String getCurrentDocument()
    {
	return currentDocument;
    }

    public  Hashtable getVisited()
    {
	return visited;
    }

    public  Hashtable getToVisit()
    {
	return toVisit;
    }

    public  int getLoadedImagesCount()
    {
	return loadedImagesCount;
    }
    

// ******************************************                          Atributes

    int depth;
    Hashtable visited = new Hashtable(); // The Keys are the names, the values, 
                                                  // the number of images found there.
    Hashtable toVisit = new Hashtable(); // The keys are the names, the values
                                                  // their depth.
    Vector keywords = new Vector();
    int loadedImagesCount = 0;
    int imagesOnThisDocument;
    int currentLevel;
    String currentDocument = "";
    String targetDir = "";
    
    Vector observers = new Vector();
    static final String DEFAULT_PARSER_NAME = "dom.wrappers.DOMParser";
}
 









